/*********************************************************************
* DCF77-Uhr als TWI-Slave (TWI via USI)
*
* Uwe Berger; 2008
*
**********************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#include "USI_TWI_Slave.h"

// TWI-Slave-Adresse
#define TWI_SLAVE_ADDR	0x10				

// Preload-Wert des Timers 
#define PRELOAD	6		// bei Prescaler 1:8 und F_CPU=1MHz sind das 500Hz

// Tiny-Board LED
#define LED_DDR  DDRB
#define LED_PORT PORTB
#define LED_PIN  PB3

// Tiny-Board INT-Ausgang
#define INT_DDR  DDRB
#define INT_PORT PORTB
#define INT_PIN  PB1

// Tiny-Board DCF77-Eingang
#define DCF_DDR  DDRB
#define DCF_PORT PORTB
#define DCF_PIN  PB4
#define DCF_INT  PCINT4

// DCF-Signal (Werte in 2ms-Werten )
#define LOW_MIN		35
#define LOW_MAX		55
#define HIGH_MIN	85
#define HIGH_MAX	105
#define ONE_SECOND	600 // mit kleiner Sicherheit; 500 waere zu knapp

// um Speicherplatz zu sparen, liegen die DCF77-Bits in einem 8-Byte-Feld (statt 60 Byte...)
// und gleich noch eine Struktur drueber, die das Auslesen der jeweiligen Werte vereinfacht,
// der dann noch frei Platz wird fuer ein paar Merker verwendet
volatile union dcf_t {
			struct	{
					uint8_t b[8];
				} bits;
			struct	{
					unsigned start_mm  : 1;	// Bit 0
					unsigned meteotime : 14;// Bit 1...14
					unsigned ruf_bit   : 1;	// Bit 15
					unsigned mez2mesz  : 1;	// Bit 16
					unsigned mez	   : 1;	// Bit 17
					unsigned mesz	   : 1;	// Bit 18
					unsigned switch_ss : 1;	// Bit 19
					unsigned begin_time: 1; // Bit 20
					unsigned mm	   : 7;	// Bit 21...27
					unsigned p_mm	   : 1;	// Bit 28
					unsigned hh	   : 6;	// Bit 29...34
					unsigned p_hh	   : 1;	// Bit 35
					unsigned dd	   : 6;	// Bit 36...41
					unsigned wd	   : 3;	// Bit 42...44
					unsigned mt	   : 5; // Bit 45...49
					unsigned yy	   : 8;	// Bit 50...57
					unsigned p_date	   : 1; // Bit 58
					unsigned click     : 1; // Bit 59 (Merker, ob neue Sekunde)
					unsigned valid     : 2; // Bit 60...61 ("Zeitermittlungsgrad" 2=OK)
					unsigned reserve   : 2;	// 2 Bit Reserve
				} dat;
			} dcf_temp;


// valide Zeit der Vorminute
typedef struct {uint8_t ss, mm, hh, dd, mt, yy, wd, mez, mesz;} date_time_t;
volatile date_time_t dcf77;

// ein paar Zaehler
volatile uint8_t peak_time = 0;		// Zeitzahler fuer Laenge eines Impulses
volatile uint8_t bit_idx = 0;		// Laufindex fuer DCF77-Bit-Feld
volatile int peak_to_peak_time = 0;     // Zeitzaehler fuer Dauer zwischen den "Beginn-Flanken"


//**************************************************************************************
// Kovertierung BCD -> Dezimal
uint8_t bcd2dec(uint8_t val)
{
	return ((val) >> 4) * 10 + ((val) & 0xf);
}


//**************************************************************************************
// alle DCF-Bits auf Null setzen
void reset_dcf_bits (void)
{
	for (uint8_t i = 0; i < 7; i++) dcf_temp.bits.b[i] = 0;
}


//**************************************************************************************
// adressiertes DCF-Bit auf 1 setzen
void set_dcf_bit (uint8_t idx)
{
	dcf_temp.bits.b[(idx/8)] |= 1 << (idx % 8);
}


//**************************************************************************************
// Datum/Uhrzeit berechnen
void generate_dcf_date_time(void)
{
	dcf77.mm   = bcd2dec(dcf_temp.dat.mm);
	dcf77.hh   = bcd2dec(dcf_temp.dat.hh);
	dcf77.dd   = bcd2dec(dcf_temp.dat.dd);
	dcf77.mt   = bcd2dec(dcf_temp.dat.mt);
	dcf77.yy   = bcd2dec(dcf_temp.dat.yy);
	dcf77.wd   = bcd2dec(dcf_temp.dat.wd);
	dcf77.mez  = dcf_temp.dat.mez;
	dcf77.mesz = dcf_temp.dat.mesz;
}


//*********************************************************************
// Datum/Zeit zuruecksetzen
void reset_dcf77_data(void)
{
	dcf77.ss   = 0;
	dcf77.mm   = 0;
	dcf77.hh   = 0;
	dcf77.dd   = 0;
	dcf77.mt   = 0;
	dcf77.yy   = 0;
	dcf77.wd   = 0;
	dcf77.mez  = 0;
	dcf77.mesz = 0;
}  

  
//*********************************************************************
// Timer
ISR(TIM0_OVF_vect)
{
	// Preloadwert setzen
	TCNT0 = PRELOAD;	
	// diverse Zeitzaehler hochzaehlen
	peak_time++;
	peak_to_peak_time++;
}   

   
//*********************************************************************
// Interruptroutine DCF-Signal
ISR(PCINT0_vect)
{
	// High -> Low ist der Impulsbeginn!!!
	if (!(PINB & ( 1 << DCF_PIN))) {
		peak_time = 0;				// Impuls beginnt
		peak_to_peak_time = 0;			// wird fuer Berechnung der 59.Sek. benoetigt
		LED_PORT |= (1 << LED_PIN);		// Status-LED ein
	} else {
		LED_PORT &= ~(1 << LED_PIN);		// Status-LED aus
		dcf_temp.dat.click = 1;			// es waere dann ein neues Bit da...
	}
}

   
//*********************************************************************
//*********************************************************************
int main(void)
{
	uint8_t temp;

	// LED- und INT-Pin als Ausgaenge
	DDRB  	 |= (1 << LED_PIN) | (1 << INT_PIN);	// LED- und INT-Pin als Ausgang schalten
	LED_PORT &= ~(1 << LED_PIN);			// LED ausschalten
	INT_PORT &= ~(1 << INT_PIN);			// INT-Port auf LOW
	
	// Timer-Overflow-Interrupt setzen
	TCNT0   = PRELOAD;				// Preloadwert setzen
	TCCR0B |= (1 << CS01);				// Prescaler Timer0 auf 8 stellen
	TIMSK  |= (1 << TOIE0);				// Timer 1 Overflow Interrupt enable

	// PCINT fuer DCF_PIN initialisieren
	DCF_PORT &= ~(1 << DCF_PIN);			// DCF-Pin als Eingang
	PCMSK    |=  (1 << DCF_INT);			// DCF-Pin maskieren
	GIMSK    |=  (1 << PCIE);			// Port Change Interrupt frei
  
	// TWI initialisieren und Slave-Adresse setzen
	USI_TWI_Slave_Initialise( TWI_SLAVE_ADDR );

	sei();
	
	reset_dcf77_data();

	for(;;)	{

		//*** TWI-Verarbeitung
		if(USI_TWI_Data_In_Receive_Buffer()) {

			// ein Byte vom Master empfangen 
			temp = USI_TWI_Receive_Byte();

			// und die Werte senden
			USI_TWI_Transmit_Byte(dcf77.ss);
			USI_TWI_Transmit_Byte(dcf77.mm);
			USI_TWI_Transmit_Byte(dcf77.hh);
			USI_TWI_Transmit_Byte(dcf77.dd);
			USI_TWI_Transmit_Byte(dcf77.mt);
			USI_TWI_Transmit_Byte(dcf77.yy);
			USI_TWI_Transmit_Byte(dcf77.wd);
			USI_TWI_Transmit_Byte(dcf77.mez);
			USI_TWI_Transmit_Byte(dcf77.mesz);
		}
		
		//*** DCF77-Verarbeitung
		// die 59. Sekunde behandeln (hier fehlt der Impuls im DCF77-Signal)
		if (peak_to_peak_time > ONE_SECOND) {
			// Sekunde "manuell" hochzaehlen
			dcf77.ss++;
			// "Zeitermittlungsgrad" eine Stufe hoeher
			if (dcf_temp.dat.valid<2) dcf_temp.dat.valid++;
			// und einmal schummeln
			dcf_temp.dat.click = 1;
			peak_to_peak_time = 0;
			bit_idx = 59;
		}

		// ein DCF77-Impuls-Ende wurde erkannt
		if (dcf_temp.dat.click) {
			
			// der aktuelle Index des Puffers ist gleichzeitig die Sekunde
			dcf77.ss = bit_idx;

			// neue Minute hat angefangen
			if (bit_idx == 0) {
				// ermittelte Daten berechnen/sichern
				if (dcf_temp.dat.valid==2) generate_dcf_date_time();
				// Puffer neu initialisieren
				reset_dcf_bits();
			}

			// der Bit-Index ist gleichzeitig die aktuelle Sekunde, ausser in der 59.
			if (bit_idx < 59) {
				// wenn eine 1 empfangen, setze entsprechendes Bit				
				if ((peak_time >= HIGH_MIN)&&(peak_time <= HIGH_MAX)) {
					set_dcf_bit(bit_idx);
				} else {
					// wenigsten ueberpruefen, ob es ein regulaerer LOW-Impuls
					// war, wenn nein, dann ungueltiges Signal und Daten-Reset...
					if (!((peak_time >= LOW_MIN)&&(peak_time <= LOW_MAX))) {
						reset_dcf_bits();
						reset_dcf77_data();
						dcf_temp.dat.valid = 0;
					}
				}
			} 

			// Bitfeldindex	entsprechend setzen		
			bit_idx++;
			if (bit_idx>58) bit_idx = 0;
			
			// Impuls auf INT-Pin erzeugen, also 2x invertieren
			INT_PORT ^= (1 << INT_PIN);     
			INT_PORT ^= (1 << INT_PIN);   

			dcf_temp.dat.click = 0;
		}
	}
}

